/*****************************************************************************

ssa_cps_info

Laurent de Soras, 2009-2010

Donne des informations sur le debit de caracteres d'un fichier
de sous-titres SSA (SubStation Alpha) ou ASS.

	---

This program is free software. It comes without any warranty, to
the extent permitted by applicable law. You can redistribute it
and/or modify it under the terms of the Do What The Fuck You Want
To Public License, Version 2, as published by Sam Hocevar. See
http://sam.zoy.org/wtfpl/COPYING for more details.

*****************************************************************************/



#include	<algorithm>
#include	<list>
#include	<string>
#include	<vector>

#include	<cassert>
#include	<cctype>
#include	<cfloat>
#include	<cmath>
#include	<cstdio>
#include	<cstdlib>



enum ReadingSpeed
{
	ReadingSpeed_TOO_SLOW = 0,
	ReadingSpeed_SLOW_ACCEPTABLE,
	ReadingSpeed_BIT_SLOW,
	ReadingSpeed_GOOD_SLOW,
	ReadingSpeed_PEFECT,
	ReadingSpeed_GOOD_FAST,
	ReadingSpeed_BIT_FAST,
	ReadingSpeed_FAST_ACCEPTABLE,
	ReadingSpeed_TOO_FAST,

	ReadingSpeed_NBR_ELT
};

static const double	reading_speed_threshold = 0.5;
static const double	reading_speed_ideal_cps = 19.0;

static const double	reading_speed_max_val [ReadingSpeed_NBR_ELT] =
{
	5, 10, 13, 15, 23, 27, 31, 35, DBL_MAX
};

static const char *	reading_speed_txt_short_0 [ReadingSpeed_NBR_ELT] =
{
	"TS", "SA", "BS", "GS", "P", "GF", "BF", "FA", "TF"
};

static const char *	reading_speed_txt_long_0 [ReadingSpeed_NBR_ELT] =
{
	"Too slow",
	"Slow, acceptable",
	"A bit slow",
	"Good, slow",
	"Perfect",
	"Good, fast",
	"A bit fast",
	"Fast, acceptable",
	"Too fast"
};

static void	compute_reading_speed (ReadingSpeed &rs, double &rs_val, int nbr_char, double duration)
{
	rs = ReadingSpeed_TOO_FAST;
	const double	rs_val_max = reading_speed_max_val [ReadingSpeed_TOO_FAST - 1];
	rs_val = rs_val_max;

	const double	margin = 0.001;
	if (duration > reading_speed_threshold + margin)
	{
		rs_val = nbr_char / (duration - reading_speed_threshold);
		while (rs_val < reading_speed_max_val [rs - 1] && rs > 0)
		{
			rs = ReadingSpeed (rs - 1);
		}
	}

	if (rs_val > rs_val_max)
	{
		rs_val = rs_val_max;
	}
}

static ReadingSpeed	compute_reading_speed (int nbr_char, double duration)
{
	double			rs_val;
	ReadingSpeed	rs;
	compute_reading_speed (rs, rs_val, nbr_char, duration);

	return (rs);
}



static int	compute_ideal_len (double duration)
{
	double		nbr_chr = 0;
	const double	margin = 0.001;
	if (duration > reading_speed_threshold + margin)
	{
		nbr_chr = (duration - reading_speed_threshold) * reading_speed_ideal_cps;
	}

	return (static_cast <int> (floor (nbr_chr + 0.5)));
}



template <class T>
static T &	access_elt (std::list <T> &l, long pos)
{
	typename std::list <T>::iterator	it = l.begin ();
	while (pos > 0)
	{
		++ it;
		-- pos;
	}

	return (*it);
}



static void	split_text (std::list <std::string> &text, const std::string &content, char delimiter, long max_nbr_strings)
{
	text.clear ();

	std::string::size_type	pos = 0;
	long				nbr_lines = 0;
	bool				cont_flag = true;
	do
	{
		const std::string::size_type	old_pos = pos;
		pos = content.find (delimiter, old_pos);
		if (pos == std::string::npos || nbr_lines >= max_nbr_strings - 1)
		{
			pos = content.length ();
			cont_flag = false;
		}

		const std::string	sub = content.substr (old_pos, pos - old_pos);
		text.push_back (sub);
		++ nbr_lines;

		++ pos;	// Skips delimiter
	}
	while (cont_flag);
}



static bool	is_valid_utf8_lead_byte (char c)
{
	const int		val = static_cast <unsigned char> (c);

	return (val < 0x80 || (val >= 0xC2 && val <= 0xF4));
}



class Date
{
public:
	int				init (const std::string &str);
	void				print () const;
	double			compute_sec () const;
	bool				operator < (const Date &other) const;

	int				_hour;
	int				_min;
	double			_sec;
};



int	Date::init (const std::string &str)
{
	int				ret_val = 0;

	std::string::size_type	delim_1 = str.find (':');
	if (delim_1 == std::string::npos)
	{
		printf ("Format de date invalide: %s\n", str.c_str ());
		ret_val = -1;
	}
	else
	{
		std::string::size_type	delim_2 = str.find (':', delim_1 + 1);

		_hour = strtol (str.c_str (), 0, 10);
		_min = strtol (str.c_str () + delim_1 + 1, 0, 10);
		_sec = strtod (str.c_str () + delim_2 + 1, 0);

		assert (compute_sec () >= 0);
	}

	return (ret_val);
}



void	Date::print () const
{
	printf ("%01d:%02d:%05.2f", _hour, _min, _sec);
}



double	Date::compute_sec () const
{
	return ((_hour * 60 + _min) * 60 + _sec);
}



bool	Date::operator < (const Date &other) const
{
	assert (&other != 0);

	return (compute_sec () < other.compute_sec ());
}



class Line
{
public:

	int				init (const std::string &raw_line, bool utf_8_flag);
	void				compute_expected_nbr_char ();
	double			get_duration () const;

	bool				operator < (const Line &other) const;

	// Source elements
	Date				_time_arr [2];			// [begin|end].
	std::string		_txt;

	// Computed things
	int				_nbr_char_arr [2];	// [measured|expected]
	int				_max_line_size;

private:

	static int		extract_date (double &t, const std::string &str);
};



typedef	std::vector <Line>	LineArray;



int	Line::init (const std::string &raw_line, bool utf_8_flag)
{
	assert (&raw_line != 0);

	int				ret_val = 0;

	std::list <std::string>	field_list;
	split_text (field_list, raw_line, ',', 10);

	if (ret_val == 0)
	{
		ret_val = _time_arr [0].init (access_elt (field_list, 1));
	}
	if (ret_val == 0)
	{
		ret_val = _time_arr [1].init (access_elt (field_list, 2));
	}

	if (ret_val == 0)
	{
		assert (! (_time_arr [1] < _time_arr [0]));

		_txt = access_elt (field_list, 9);

		enum State
		{
			State_NORMAL = 0,
			State_ESC_CHAR,
			State_ESC_SEQ
		};

		_max_line_size = 0;
		long				str_real_len = 0;
		long				cur_line_size = 0;
		long				nbr_pot_ws_at_the_eol = 0;
		bool				line_started_flag = false;	// To strip whitespace at the beginning
		State				state = State_NORMAL;
		for (std::string::size_type pos = 0; pos < _txt.length (); ++pos)
		{
			char				c = _txt [pos];
			bool				char_flag = false;

			switch (state)
			{
			case	State_NORMAL:
				if (c == '\\')
				{
					state = State_ESC_CHAR;
				}
				else if (c == '{')
				{
					state = State_ESC_SEQ;
				}
				else
				{
					char_flag = true;
				}
				break;

			case	State_ESC_CHAR:
				if (c != 'N' && c != 'n')
				{
					if (c == 'h')
					{
						c = ' ';	// To be considered later as a whitespace
					}
					else
					{
						++ cur_line_size;
					}
					char_flag = true;
				}
				else if (c == 'n')
				{
					c = ' ';	// To be considered later as a whitespace
				}
				else if (c == 'N')
				{
					cur_line_size -= nbr_pot_ws_at_the_eol;
					assert (cur_line_size >= 0);
					_max_line_size =
						  (_max_line_size < cur_line_size)
						? cur_line_size
						: _max_line_size;
					str_real_len += cur_line_size;
					cur_line_size = 0;
					nbr_pot_ws_at_the_eol = 0;
					line_started_flag = false;
				}
				state = State_NORMAL;
				break;

			case	State_ESC_SEQ:
				if (c == '}')
				{
					state = State_NORMAL;
				}
				break;
			}

			if (char_flag)
			{
				if (! utf_8_flag || is_valid_utf8_lead_byte (c))
				{
					if (isspace (c))
					{
						if (line_started_flag)
						{
							++ nbr_pot_ws_at_the_eol;
						}
					}
					else
					{
						nbr_pot_ws_at_the_eol = 0;
						line_started_flag = true;
					}

					if (line_started_flag)
					{
						++ cur_line_size;
					}
				}
			}
		}

		cur_line_size -= nbr_pot_ws_at_the_eol;
		assert (cur_line_size >= 0);
		_max_line_size =
			  (_max_line_size < cur_line_size)
			? cur_line_size
			: _max_line_size;
		str_real_len += cur_line_size;
		cur_line_size = 0;

		_nbr_char_arr [0] = str_real_len;
	}

	return (ret_val);
}



void	Line::compute_expected_nbr_char ()
{
	const double	duration = get_duration ();
	const int		nbr_char = compute_ideal_len (duration);
	_nbr_char_arr [1] = nbr_char;
}



double	Line::get_duration () const
{
	return (_time_arr [1].compute_sec () - _time_arr [0].compute_sec ());
}



bool	Line::operator < (const Line &other) const
{
	assert (&other != 0);

	return (_time_arr [0] < other._time_arr [0]);
}



static int	read_file (std::string &content, const char *filename_0)
{
	int				ret_val = 0;

	content.erase (content.begin (), content.end ());

	FILE *			f_ptr = fopen (filename_0, "ra");
	if (f_ptr == 0)
	{
		printf ("Impossible d\'ouvrir le fichier %s\n", filename_0);
		ret_val = -1;
	}

	if (ret_val == 0)
	{
		int				c;
		do
		{
			c = fgetc (f_ptr);
			if (c != EOF)
			{
				content += c;
			}
		}
		while (c != EOF);
	}

	if (f_ptr != 0)
	{
		fclose (f_ptr);
		f_ptr = 0;
	}

	return (ret_val);
}



// Removes only empty lines and whitespaces at the beginning of the lines
static void	remove_white (std::list <std::string> &text)
{
	typedef	std::list <std::string>::iterator	ListIt;
	ListIt		it = text.begin ();
	while (it != text.end ())
	{
		ListIt			it_cur = it;
		++ it;

		std::string &	str = *it_cur;
		while (str.length () > 0 && isspace (str [0]))
		{
			str.erase (str.begin ());
		}

		if (str.length () == 0)
		{
			text.erase (it_cur);
		}
	}
}



// Starts from it, returns found element in it.
// Sets it at text.end() if the label is not found.
static void	find_label (std::list <std::string>::const_iterator &it, const std::list <std::string> &text, const char *label_0)
{
	bool				found_flag = false;
	while (! found_flag && it != text.end ())
	{
		if (it->find (label_0) == 0)
		{
			found_flag = true;
		}

		else
		{
			++ it;
		}
	}
}



static void	collect_lines_beginning_with (const char label_0 [], std::list <std::string> &text_new, const std::list <std::string> &text, std::list <std::string>::const_iterator it_beg, std::list <std::string>::const_iterator it_end)
{
	text_new.clear ();

	std::list <std::string>::const_iterator	it = it_beg;
	while (it != it_end)
	{
		if (it->find (label_0) == 0)
		{
			text_new.push_back (*it);
		}

		++ it;
	}
}



// Based only on BOM (Byte Order Mark)
static bool	check_utf_8 (const std::string &str)
{
	return (   str.length () >= 3
	        && str [0] == '\xEF'
	        && str [1] == '\xBB'
	        && str [2] == '\xBF');
}



static int	extract_lines (LineArray &line_arr, const std::string &content, bool utf_8_flag)
{
	int				ret_val = 0;

	line_arr.clear ();

	std::list <std::string>	text;
	split_text (text, content, '\n', LONG_MAX);
	remove_white (text);

	// Finds boundaries
	std::list <std::string>::const_iterator	it_beg = text.begin ();
	find_label (it_beg, text, "[Events]");

	std::list <std::string>::const_iterator	it_end = it_beg;
	++ it_end;
	find_label (it_end, text, "[");

	// Removes useless lines
	std::list <std::string>	text_tmp;
	collect_lines_beginning_with ("Dialogue:", text_tmp, text, it_beg, it_end);
	text = text_tmp;

	// Collects line data
	it_beg = text.begin ();
	it_end = text.end ();
	for (std::list <std::string>::const_iterator it = it_beg; it != it_end; ++it)
	{
		Line				line;
		int				ret_val = line.init (*it, utf_8_flag);
		if (ret_val == 0)
		{
			line_arr.push_back (line);
		}
	}

	return (ret_val);
}



int	main (int argc, char *argv [])
{
	int				ret_val = 0;

	int				max_len = 40;		// Characters
	double			min_dur = 0.8;		// s
	const double	overlap_magin = 0.15;	// s

	if (argc < 2 || argc > 2)
	{
		printf (
			"Usage:\n\n"
			"ssa_cps_info <filename>\n\n"
			"    filename    Nom du fichier .ssa ou .ass a analyser.\n"
			"\n"
		);
		ret_val = -1;
	}

	std::string		content;
	if (ret_val == 0)
	{
		ret_val = read_file (content, argv [1]);
	}

	LineArray		line_arr;
	if (ret_val == 0)
	{
		const bool		utf_8_flag = check_utf_8 (content);
		ret_val = extract_lines (line_arr, content, utf_8_flag);
	}

	std::vector <int>	nbr_st_per_rs (ReadingSpeed_NBR_ELT, 0);
	int				nbr_valid_st = 0;
	long				st_len_tot = 0;
	int				nbr_overlaps = 0;

	double			rs_val_tot = 0;
	int				nbr_st_rs_acceptable = 0;

	if (ret_val == 0)
	{
		std::sort (line_arr.begin (), line_arr.end ());

		for (LineArray::size_type index = 0; index < line_arr.size (); ++index)
		{
			Line &			line = line_arr [index];
			line.compute_expected_nbr_char ();

			bool				overlap_flag = false;
			if (index > 0)
			{
				const Line &	prev_line = line_arr [index - 1];
				const double	prev_end_date = prev_line._time_arr [1].compute_sec ();
				const double	cur_beg_date = line._time_arr [0].compute_sec ();
				const double	dist = cur_beg_date - prev_end_date;
				if (dist < overlap_magin)
				{
					overlap_flag = true;
					++ nbr_overlaps;
				}
			}

			const int		nbr_real = line._nbr_char_arr [0];
			const int		nbr_exp = line._nbr_char_arr [1];
			const double	duration = line.get_duration ();

			ReadingSpeed	rs;
			double			rs_val;
			compute_reading_speed (rs, rs_val, nbr_real, duration);
			if (nbr_real > 0)
			{
				st_len_tot += nbr_real;
				++ nbr_st_per_rs [rs];
				++ nbr_valid_st;

				if (rs > ReadingSpeed_TOO_SLOW && rs < ReadingSpeed_TOO_FAST)
				{
					rs_val_tot += rs_val;
					++ nbr_st_rs_acceptable;
				}
			}

			const bool		ok_flag = (   rs > ReadingSpeed_TOO_SLOW
								           && rs < ReadingSpeed_FAST_ACCEPTABLE);
			if (ok_flag || nbr_real == 0)
			{
				printf ("  ");
			}
			else
			{
				const char *	rs_txt_0 = reading_speed_txt_short_0 [rs];
				printf ("%2s", rs_txt_0);
			}

			printf (
				" %s%s",
				(duration + 1e-4 < min_dur) ? "D" : " ",	// + small value because of rounding errors
				(overlap_flag)              ? "-" : " "
			);

			if (line._max_line_size <= max_len)
			{
				printf ("   ");
			}
			else
			{
				printf (" %2d", line._max_line_size);
			}

			printf (" date: ");
			line._time_arr [0].print ();

			printf (", duree: %5.2f", duration);

			printf (
				", signes: %3d, attendu: %3d, texte: %s\n",
				nbr_real,
				nbr_exp,
				line._txt.c_str ()
			);
		}
	}

	if (ret_val == 0)
	{
		printf ("\nStatistiques :\n");

		static const char	bar_0 [] = "########################################";
		static const int	bar_len = sizeof (bar_0) - 1;

		for (int rs = 0; rs < ReadingSpeed_NBR_ELT; ++rs)
		{
			const int		nbr_st = nbr_st_per_rs [rs];
			const char *	rs_txt_0 = reading_speed_txt_long_0 [rs];
			const double	ratio = double (nbr_st) / double (nbr_valid_st);
			assert (ratio >= 0);
			assert (ratio <= 1);

			const double	percent = ratio * 100;
			printf ("%4d %-20s (%5.1f %%) ", nbr_st, rs_txt_0, percent);

			const int		bar_pos = int (floor (ratio * bar_len + 0.5));
			printf ("%s\n", bar_0 + bar_len - bar_pos);
		}

		{
			const double	ratio = double (nbr_overlaps) / double (nbr_valid_st);
			const double	percent = ratio * 100;
			printf ("%4d superpositions       (%5.1f %%)\n", nbr_overlaps, percent);
		}

		const double	avg_rs_acceptable = rs_val_tot / nbr_st_rs_acceptable;
		const double	avg_st_len = double (st_len_tot) / double (nbr_valid_st);
		printf ("Vitesse globale de lecture     : %5.1f c/s\n", avg_rs_acceptable);
		printf ("Taille moyenne d\'un sous-titre : %5.1f c\n", avg_st_len);

		printf ("\n");
	}

	if (ret_val != 0)
	{
		printf ("Erreur.\n\n");
	}

	fgetc (stdin);

	return (ret_val);
}
